@@ -1,27 +1,32 @@
+"""
+ExtDirect implementation for Pyramid
+"""
 from collections import defaultdict
 import json
+import logging
 import traceback
+try:
+    from html.entities import entitydefs  # Python 3
+except ImportError:
+    from htmlentitydefs import entitydefs  # Python 2
 
-from pyramid.security import has_permission
 from pyramid.view import render_view_to_response
 from webob import Response
 from zope.interface import implementer
 from zope.interface import Interface
 import venusian
 
-try:
-    from html.entities import entitydefs  # Python 3
-except ImportError:
-    from htmlentitydefs import entitydefs  # Python 2
+__version__ = '0.6.1'
 
-__version__ = "0.3.2"
 
+LOG = logging.getLogger(__name__)
 
 # form parameters sent by ExtDirect when using a form-submit
 # see http://www.sencha.com/products/js/direct.php
 FORM_DATA_KEYS = frozenset([
     "extAction",
     "extMethod",
+    "extMetadata",
     "extTID",
     "extUpload",
     "extType"
@@ -29,12 +34,17 @@ FORM_DATA_KEYS = frozenset([
 
 # response to a file upload cannot be return as application/json, ExtDirect
 # defines a special html response body for this use case where the response
-# data is added to a textarea for faster JS-side decoding (since etxtarea text
+# data is added to a textarea for faster JS-side decoding (since textarea text
 # is not a DOM node)
-FORM_SUBMIT_RESPONSE_TPL = '<html><body><textarea>%s</textarea></body></html>'
+FORM_SUBMIT_RESPONSE_TPL = '<html><body><textarea>{}</textarea></body></html>'
+
+JS_API_TPL = """var Ext = Ext || {{}};
+Ext.ns('{namespace}'); {descriptor} = {api};
+"""
 
 
 def _mk_cb_key(action_name, method_name):
+    """ helper function to create a unique actions dict key """
     return action_name + '#' + method_name
 
 
@@ -45,10 +55,10 @@ class JsonReprEncoder(json.JSONEncoder):
             # return decoded response body in case it's an already
             # rendered exception view
             return json.loads(obj.unicode_body)
-        jr = getattr(obj, 'json_repr', None)
-        if jr is None:
+        json_repr = getattr(obj, 'json_repr', None)
+        if json_repr is None:
             return super(JsonReprEncoder, self).default(obj)
-        return jr()
+        return json_repr()
 
 
 class IExtdirect(Interface):
@@ -63,7 +73,7 @@ class AccessDeniedException(Exception):
 @implementer(IExtdirect)
 class Extdirect(object):
     """
-    Handles ExtDirect API respresentation and routing.
+    Handles ExtDirect API representation and routing.
 
     The Extdirect accepts a number of arguments: ``app``,
     ``api_path``, ``router_path``, ``namespace``, ``descriptor``
@@ -86,7 +96,7 @@ class Extdirect(object):
 
     If ``expose_exceptions`` argument is set to True the exception
     traceback will be exposed in the response object. WARNING this
-    is potentially dangeerous, do not use in production environments.
+    is potentially dangerous, do not use in production environments.
 
     The ``debug_mode`` argument will create a 'message' key in the
     response object pointing to a structure that can be used in pyramid
@@ -100,10 +110,12 @@ class Extdirect(object):
     """
 
     def __init__(self, api_path="extdirect-api.js",
-                 router_path="extdirect-router", namespace='Ext.app',
+                 router_path="extdirect-router",
+                 namespace='Ext.app',
                  descriptor='Ext.app.REMOTING_API',
                  expose_exceptions=True,
-                 debug_mode=False):
+                 debug_mode=False,
+                 json_encoder=JsonReprEncoder):
         self.api_path = api_path
         self.router_path = router_path
         self.namespace = namespace
@@ -111,6 +123,7 @@ class Extdirect(object):
         self.expose_exceptions = expose_exceptions
         self.debug_mode = debug_mode
         self.actions = defaultdict(dict)
+        self.json_encoder = json_encoder
 
     def add_action(self, action_name, **settings):
         """
@@ -125,6 +138,7 @@ class Extdirect(object):
         ``numargs``: Number of arguments passed to the wrapped callable
         ``accept_files``: If true, this action will be declared as formHandler in API
         ``permission``: The permission needed to execute the wrapped callable
+        ``metadata``: Metadata definition
         ``request_as_last_param``: If true, the wrapped callable will receive a request object
             as last argument
 
@@ -135,26 +149,38 @@ class Extdirect(object):
     def get_actions(self):
         """ Builds and returns a dict of actions to be used in ExtDirect API """
         ret = {}
-        for (k, v) in self.actions.items():
+        for (key, val) in self.actions.items():
             items = []
-            for settings in v.values():
-                d = dict(
-                    len = settings['numargs'],
-                    name = settings['method_name']
+            for settings in val.values():
+                method_info = dict(
+                    len=settings['numargs'],
+                    name=settings['method_name']
                 )
                 if settings['accepts_files']:
-                    d['formHandler'] = True
-                items.append(d)
-            ret[k] = items
+                    method_info['formHandler'] = True
+                meta = settings['metadata']
+                if meta:
+                    if isinstance(meta, ExtListMetadata):
+                        method_info['metadata'] = {
+                            'len': meta.numargs,
+                            'strict': meta.strict
+                        }
+                    elif isinstance(meta, ExtDictMetadata):
+                        method_info['metadata'] = {
+                            'params': meta.param_names,
+                            'strict': meta.strict
+                        }
+                items.append(method_info)
+            ret[key] = items
         return ret
 
     def get_method(self, action, method):
         """ Returns a method's settings """
         if action not in self.actions:
-            raise KeyError("Invalid action: " + action)
+            raise KeyError("Invalid action: {}".format(action))
         key = _mk_cb_key(action, method)
         if key not in self.actions[action]:
-            raise KeyError("No such method in '%s': '%s':" % (action, method))
+            raise KeyError("No such method in '{}': '{}'".format(action, method))
         return self.actions[action][key]
 
     def _get_api_dict(self, request):
@@ -168,19 +194,23 @@ class Extdirect(object):
                     actions[action] = all_actions[action]
         else:
             actions = all_actions
+        # FIXME: use route_url instead of request.application_url + '/' + self.router_path,
         return dict(
-            url = request.application_url + '/' + self.router_path,
-            type = 'remoting',
-            namespace = self.namespace,
-            actions = actions
+            url=request.application_url + '/' + self.router_path,
+            type='remoting',
+            namespace=self.namespace,
+            actions=actions
         )
 
     def dump_api(self, request):
         """ Dumps all known remote methods """
-        return """Ext.ns('%s'); %s = %s;""" % \
-            (self.namespace, self.descriptor, json.dumps(self._get_api_dict(request)))
+        return JS_API_TPL.format(
+            namespace=self.namespace,
+            descriptor=self.descriptor,
+            api=json.dumps(self._get_api_dict(request))
+        )
 
-    def _do_route(self, action_name, method_name, params, trans_id, request):
+    def _do_route(self, action_name, method_name, params, metadata, trans_id, request):
         """ Performs routing, i.e. calls decorated methods/functions """
         if params is None:
             params = list()
@@ -200,38 +230,48 @@ class Extdirect(object):
         permission_ok = True
         context = request.root
 
+        prepend = []
         if settings['class']:
             instance = settings['class'](request)
-            params.insert(0, instance)
+            prepend.append(instance)
             context = instance
         elif append_request:
             params.append(request)
 
+        if settings['metadata']:
+            prepend.append(metadata)
+
+        params = prepend + params
+
         if permission is not None:
-            permission_ok = has_permission(permission, context, request)
+            permission_ok = request.has_permission(permission, context)
 
         try:
             if not permission_ok:
                 raise AccessDeniedException("Access denied")
             ret["result"] = callback(*params)
-        except Exception as e:
+        except Exception as exc:
             ret["type"] = "exception"
             # Let a user defined view for specific exception prevent returning
             # a server error.
-            exception_view = render_view_to_response(e, request)
+            exception_view = render_view_to_response(exc, request)
             if exception_view is not None:
                 ret["result"] = exception_view
                 return ret
 
+            # Log Error
+            LOG.error("%s: %s", str(exc.__class__.__name__), exc)
+            LOG.info(traceback.format_exc())
+
             if self.expose_exceptions:
                 ret["result"] = {
                     'error': True,
-                    'message': str(e),
-                    'exception_class': str(e.__class__),
+                    'message': str(exc),
+                    'exception_class': str(exc.__class__),
                     'stacktrace': traceback.format_exc()
                 }
             else:
-                message = 'Error executing %s.%s' % (action_name, method_name)
+                message = 'Error executing {}.{}'.format(action_name, method_name)
                 ret["result"] = {
                     'error': True,
                     'message': message
@@ -245,50 +285,92 @@ class Extdirect(object):
                 import sys
                 exc_history = request.exc_history
                 if exc_history is not None:
-                    tb = get_traceback(info=sys.exc_info(),
-                            skip=1,
-                            show_hidden_frames=False,
-                            ignore_system_exceptions=True)
+                    tb = get_traceback(
+                        info=sys.exc_info(),
+                        skip=1,
+                        show_hidden_frames=False,
+                        ignore_system_exceptions=True)
                     for frame in tb.frames:
                         exc_history.frames[frame.id] = frame
                     exc_history.tracebacks[tb.id] = tb
 
                     qs = {'token': exc_history.token, 'tb': str(tb.id)}
-                    msg = 'Exception: traceback url: %s'
                     exc_url = request.route_url(EXC_ROUTE_NAME, _query=qs)
-                    exc_msg = msg % (exc_url)
-                    ret["message"] = exc_msg
+                    ret['message'] = 'Exception: traceback url: {}'.format(exc_url)
         return ret
 
     def route(self, request):
+        """ Route a request to the corresponding action method """
         is_form_data = is_form_submit(request)
         if is_form_data:
-            params = parse_extdirect_form_submit(request)
+            data = parse_extdirect_form_submit(request)
         else:
-            params = parse_extdirect_request(request)
+            data = parse_extdirect_request(request)
         ret = []
-        for (act, meth, params, tid) in params:
-            ret.append(self._do_route(act, meth, params, tid, request))
+        for (act, meth, params, metadata, tid) in data:
+            ret.append(self._do_route(act, meth, params, metadata, tid, request))
         if not is_form_data:
             if len(ret) == 1:
                 ret = ret[0]
-            return (json.dumps(ret, cls=JsonReprEncoder), False)
+            return (json.dumps(ret, cls=self.json_encoder), False)
         ret = ret[0] # form data cannot be batched
-        s = json.dumps(ret, cls=JsonReprEncoder).replace("&quot;", r"\&quot;");
-        return (FORM_SUBMIT_RESPONSE_TPL % (s,), True)
+        form_data = json.dumps(ret, cls=self.json_encoder).replace("&quot;", r"\&quot;")
+        return (FORM_SUBMIT_RESPONSE_TPL.format(form_data), True)
 
 
-class extdirect_method(object):
+class ExtMetadata(object):
+    """ Base metadata class """
+
+    def __init__(self, strict=True):
+        self.strict = strict
+
+
+class ExtListMetadata(ExtMetadata):
+    """ Represents a metadata definition that will
+        pass in an ordered list of parameters
+        ``metadata`` arguments
+    """
+
+    def __init__(self, numargs=None, strict=True):
+        super(ExtListMetadata, self).__init__(strict)
+        if type(numargs) is not int or numargs < 1:
+            raise ValueError("'numargs' has to be an int > 0")
+        self.numargs = numargs
+
+
+class ExtDictMetadata(ExtMetadata):
+    """ Represents a metadata definition that will
+        pass in a dict as ``metadata`` argument
+        with members defined in ``param_names``
+    """
+
+    def __init__(self, param_names=None, strict=True):
+        super(ExtDictMetadata, self).__init__(strict)
+        if self.strict and not param_names:
+            raise ValueError("Strict is set, please provide at least one param name")
+        self.param_names = param_names
+
+
+class extdirect_method(object): # pylint: disable=invalid-name, too-few-public-methods
     """Enables direct extjs access to python methods through json/form submit"""
-    def __init__(self, action=None, method_name=None, permission=None,
-                 accepts_files=False, request_as_last_param=False):
+    def __init__(self,
+            action=None,
+            method_name=None,
+            permission=None,
+            accepts_files=False,
+            metadata=None,
+            request_as_last_param=False):
+        if metadata and not isinstance(metadata, ExtMetadata):
+            raise ValueError("Metadata must be an instance of either ExtListMetadata or ExtDictMetadata")
+        self.info = None
         self._settings = dict(
-            action = action,
-            method_name = method_name,
-            permission = permission,
-            accepts_files = accepts_files,
-            request_as_last_param = request_as_last_param,
-            original_name = None,
+            action=action,
+            method_name=method_name,
+            permission=permission,
+            accepts_files=accepts_files,
+            metadata=metadata,
+            request_as_last_param=request_as_last_param,
+            original_name=None
         )
 
     def __call__(self, wrapped):
@@ -297,35 +379,46 @@ class extdirect_method(object):
         if self._settings["method_name"] is None:
             self._settings["method_name"] = original_name
 
-        self.info = venusian.attach(wrapped,
-                                    self.register,
-                                    category='extdirect')
+        self.info = venusian.attach(
+            wrapped,
+            self.register,
+            category='extdirect'
+        )
         return wrapped
 
     def _get_settings(self):
         return self._settings.copy()
 
-    def register(self, scanner, name, ob):
+    def register(self, scanner, name, obj):
+        """ Venusian callback that registers a function
+            or a class method as extdirect method
+        """
         settings = self._get_settings()
+        metadata = settings['metadata']
 
-        class_context = isinstance(ob, type)
+        class_context = isinstance(obj, type)
 
         settings['class'] = None
         if class_context:
-            settings['class'] = ob
-            callback = getattr(ob, settings["original_name"])
+            settings['class'] = obj
+            callback = getattr(obj, settings["original_name"])
             numargs = callback.__code__.co_argcount
             # instance var doesn't count
             numargs -= 1
         else:
             if settings['action'] is None:
                 raise ValueError("Decorated function has no action (name)")
-            callback = ob
+            callback = obj
             numargs = callback.__code__.co_argcount
 
         if numargs and settings['request_as_last_param']:
             numargs -= 1
 
+        if metadata:
+            numargs -= 1
+            if numargs < 0:
+                raise ValueError("{} must provide at least one argument for metadata".format(settings['original_name']))
+
         settings['numargs'] = numargs
 
         action = settings.pop("action", None)
@@ -333,7 +426,7 @@ class extdirect_method(object):
             name = action
 
         if class_context:
-            class_settings = getattr(ob, '__extdirect_settings__', None)
+            class_settings = getattr(obj, '__extdirect_settings__', None)
             if class_settings:
                 if action is None:
                     name = class_settings.get("default_action_name", name)
@@ -360,11 +453,14 @@ def parse_extdirect_form_submit(request):
     action = params.get('extAction')
     method = params.get('extMethod')
     tid = params.get('extTID')
+    metadata = params.get('extMetadata')
+    if metadata:
+        metadata = json.loads(metadata)
     data = dict()
     for key in params:
         if key not in FORM_DATA_KEYS:
             data[key] = params[key]
-    return [(action, method, [data], tid)]
+    return [(action, method, [data], metadata, tid)]
 
 
 def parse_extdirect_request(request):
@@ -377,12 +473,13 @@ def parse_extdirect_request(request):
     ret = []
     if not isinstance(decoded_body, list):
         decoded_body = [decoded_body]
-    for p in decoded_body:
-        action = p['action']
-        method = p['method']
-        data = p['data']
-        tid = p['tid']
-        ret.append((action, method, data, tid))
+    for part in decoded_body:
+        action = part['action']
+        method = part['method']
+        data = part['data']
+        metadata = part.get('metadata', None)
+        tid = part['tid']
+        ret.append((action, method, data, metadata, tid))
     return ret
 
 
@@ -402,16 +499,20 @@ def router_view(request):
 
 
 def includeme(config):
-    """Let extdirect be included by config.include()."""
+    """ Let extdirect be included by config.include(). """
     settings = config.registry.settings
     extdirect_config = dict()
     names = ("api_path", "router_path", "namespace", "descriptor",
-             "expose_exceptions", "debug_mode")
+             "expose_exceptions", "debug_mode", "json_encoder")
     for name in names:
-        qname = "pyramid_extdirect.%s" % name
+        qname = "pyramid_extdirect.{}".format(name)
         value = settings.get(qname, None)
         if name == "expose_exceptions" or name == "debug_mode":
             value = (value == "true")
+        if name == "json_encoder" and value:
+            from pyramid.path import DottedNameResolver
+            resolver = DottedNameResolver()
+            value = resolver.resolve(value)
         if value is not None:
             extdirect_config[name] = value
 
